#!/usr/local/bin/perl
#
#	@(#) security.pl v2.8.1 System Security script
#
# Changes
#	v 2.8.1
#		Update new SHA1 package and remove NIS.  Allow writing a log
#		file instead of mailing results. fix kern.module_path coding.
#	v 2.8.0 
#		Added code to the .cf file to list the devices root is
#		allowed to login to and added the code to this script.
#	v 2.7.1
#		Changed the code for FreeBSD checking for which ports root
#		is allowed to login to.
#	v 2.7.0
#		Added the ignoredevdir in the f file so we can ignore the
#		block special devices in /compat/*/dev too.
#	v 2.6.5
#		Changed wanted routine to ignore files that are links.
#	v 2.6.4
#		Added type autofs to ignore and prune in the wanted subroutine.
#	v 2.6.3
#		Change code to determine if non local filesystem, remove the
#		support for MD5 and sum (just stick with SHA). 
#	v 2.6.2
#		Added mtree checks for /usr/X11R6 with the BSD.x11.dist
#	v 2.6.1
#		Fixed bug where if root was not allowed on all ports, it reported
#		that it was
#	v 2.6.0
#		Add checks to see if root is allowed to login on other than console
#	v 2.5.3
#		Fixed code to pull in the YP passwd and groups on solaris systems.
#		It looked for the +:: in the files, but that is not needed under
#		solaris systems.
#	v 2.5.2
#		Added sysctl call to find the kernel module path for FreeBSD
#	v 2.5.1
#		Defined the default digest function if the .digest file does not
#		have it defined.
#	v 2.5.0
#		Changed the MD5 digests to configurable digest algorithms.
#		The supported digests are SHA, MD5, and sum.  Added checks of 
#		loadable kernel modules.
#	v 2.4.5
#		Added fixes for yp entries of '+<logname>::::::'
#	v 2.4.4
#		changed email address to "sysnotify" from "security"
#	v 2.4.3
#		Added the 2> /dev/null to showmount for hosts that are
#		not nfs servers.
#	v 2.4.2
#		Hard coded the file names to put the checksums in the MD5 file
#	v 2.4.1
#		Added code when displaying $0 it is the shortend name.  Reordered
#		the options to be in alphabetical order.
#	v 2.4.0
#		Added code to save runs of $pkgchk or $showfiles to do diffs on
#		them since there are so many instances that cannot be resolved.
#		They are the .today & .yesterday files.  Also changed /usr/bin/diff
#		to be set in the configure subroutine.  Changed $SUM to be the 
#		internal MD5 perl routine instead of a seperate program.
#	v 2.3.0
#		Added 'pkgchk -q -n' for Solaris to be the equivelant of 'mtree'
#		for FreeBSD.  Also added showfiles for SGI to do the same thing.
#		Changed the location for the setuid.* and conf files to a more 
#		secure directory.  Added an MD5 file for the all config files.
#	v 2.2.1
#		Changed the $SUM values to be absolute values, and to use MD5
#		for others than just FreeBSD.
#	v 2.2.0
#		Added code to check for intruder configuration files.
#	v 2.1.0
#		Added code to check the /etc/shadow file where existant.
#	v 2.0.0
#		Created the /etc/security.cf file to configure what should be 
#		ignored and stuff like that.  Added a 'touch' of 
#		/var/adm/setuid.yesterday, if it does not exist) so the diff 
#		will work.  Added the ($dev < 40,000,000) in wanted for Solaris 
#		2.6 (NFS filesystems seem to have $dev numbers > 40M, instead 
#		of < 0).  Added code to check for setuid/setgid programs that are
#		writeable by other than the owner. 
#	v 1.6.1
#		Added -v to $mailcmd, added $mailargs, added /usr/ucb/Mail to 
#		configure, added smpt to the list of uid 0 accounts to ignore,
#		changed from reading directly the /etc/exports|/etc/dfs/dfstab
#		to use the output from showmount -e localhost.
#	v 1.6.0
#		Added YP map calls when the '+:' token is in the local file
#	v 1.5.2
#		More disney specific stuff
#	v 1.5.1
#		Added the undef $perms{$dir} so when mtree runs fine, there is
#		no output.
#	v 1.5.0
#		Added mtree call for systems that have them (namely FreeBSD).
#	v 1.0.3
#		Added chmod/chown to the setuid.today file so that it has the
#		correct permissions
#	v 1.0.2
#		Added some checks in wanted so not to check the /proc filesystem
#	v 1.0.1 
#		Modified the configure routine to use /usr/5bin/sum if it exists

use Getopt::Std;
use Sys::Hostname;
use POSIX;
use Digest::SHA1;
use File::Find;
use Filesys::Statvfs;

# The SHA routine:
sub SHA {
	my ($file) = @_;
	my ($sha) = new Digest::SHA1;

	open (sha_fd, $file);
	$sha->reset();
	$sha->addfile(*sha_fd{IO});
	my ($hex) = $sha->hexdigest();
	close sha_fd;
	return "$sha_ver($file) = $hex\n";
}

# Check to see what hosts have filesystems exported to the world.  This 
# idea comes from the HP-UX System Security manual
sub check_exports {
	my ($dir, $dummy);

	print ( STDERR "In check_exports\n") if ( $opt_d );
	if ( -f $showmount ) {
		open(fp, "$showmount -e localhost 2>/dev/null |") || die "Could not run $showmount";
		while (<fp>) {
			chop;
			$dummy = $_;
			$dummy =~ s/^.* //o;
			$dir = $_;
			$dir =~ s/ .*$//o;
			$exports{$dir} = 1 if ($dummy =~ /\(everyone\)/o);
		}
		close fp;
	}
}

sub create_secure_file {
	my ($file, $uid, $gid, $mode) = @_;

	umask 077;
	$( = $uid; $) = $gid;
	open (FP, "> $file") || die "Could not open secure file";
	close FP;
	chown ( $uid, $gid, $file);
	chmod ( $mode, $file);
}

sub check_intruder_files {
	print ( STDERR "In check_intruder_files\n") if ( $opt_d );

	foreach $file (split (' ', $intruderfiles)) {
		$badfile{$file} = 1 if ( -f $file );
	}
}

sub check_passwd {
	print ( STDERR "In check_passwd\n") if ( $opt_d );
	if ( -f $shadow ) {
		&create_secure_file ("$CONFDIR/passwd", 0, 0, 0600);
		&create_secure_file ("$CONFDIR/shadow", 0, 0, 0600);
		`sort -t: $passwd > "$CONFDIR/passwd"`;
		`sort -t: $shadow > "$CONFDIR/shadow"`;

		open(fp, "join -a 1 -t : -o 1.1,2.2,1.3,1.4,1.5,1.6,1.7 $CONFDIR/passwd $CONFDIR/shadow |") || die "Could not read join passwd & shadow files";
	} else {
		open(fp, "sort -t: +2n $passwd |") || die "Could not read sorted passwd file";
	}

	$olduid = -5;
	while (<fp>) {
		chop;
		($Fld1,$Fld2,$Fld3,$Fld4) = split(':', $_, 9999);

		# Check to see who has uids of 0.  This comes from the security 
		# script from BSD 4.4.  We do not need to report the uid 0 on 
		# the 'root' account, nor accounts that start with '+'
		$uid0{$Fld1} = 1 
			if (! $Fld3 && 
				$Fld1 ne 'root' && 
				$Fld1 !~ /^\+/ && 
				index ($ignoreuid, $Fld1)<0);

		# Check to see who has accounts without passwords.  This comes 
		# from the security script from BSD 4.4.  Ignore the accounts
		# that start with '+'
		$nopass{$Fld1} = 1 if ($Fld2 eq '' && 
				$Fld1 !~ /^\+/ && 
				index ($ignorepasswd, $Fld1)<0);

		# Check to see what accounts have duplicate uid's (except the 
		# root account, which was handled earlier.  This idea comes from 
		# the HP-UX System Security manual
		$dupuid{$Fld1} = $oldFld1 if ( $olduid eq $Fld3 && $Fld3 );

		$olduid = $Fld3;
		$oldFld1 = $Fld1;
	}

	# Remove the tmp files.
	unlink ("$CONFDIR/shadow") && unlink ("$CONFDIR/passwd") if ( -f $shadow );
	close fp;
}

# Do a diff of the two files to make sure that no files have changed.
# The idea orginally came from FreeBSD, but I changed the output of the find
# from an ls to a sum, that idea came from the SGI System Security manuals.
sub check_setuid {
	print ( STDERR "In check_setuid\n") if ( $opt_d );

	# First rename yesterday's file if needed.
	if ( -f "$CONFDIR/setuid.today" ) {
		rename ("$CONFDIR/setuid.today", "$CONFDIR/setuid.yesterday");
	} else {
		&create_secure_file ("$CONFDIR/setuid.yesterday", 0, 0, 0600);
	}
	&create_secure_file ("$CONFDIR/setuid.today", 0, 0, 0600);

	# Now write out today's info for tomorrow
	open (FP, "> $CONFDIR/setuid.today") || die "Could not open setuid.today";
	foreach $i (sort keys %setuid) {
		print FP $setuid{$i};
	}
	close FP;
}

sub run_mtree {
	my ($dir, $dir2, $file);
	print ( STDERR "In run_mtree\n") if ( $opt_d );

	if ( opendir (DIR, '/etc/mtree')) {  
		while ($file = readdir (DIR)) {
			$dir = '';
			$dir = '/'  if ( $file =~ /BSD.root.dist/ );
			$dir = '/var'  if ( $file =~ /BSD.var.dist/ );
			$dir = '/usr'  if ( $file =~ /BSD.usr.dist/ );
			$dir = '/usr/include'  if ( $file =~ /BSD.include.dist/ );
			$dir = '/usr/X11R6'  if ( $file =~ /BSD.x11.dist/ );
			$dir = '/usr/local'  if ( $file =~ /BSD.local.dist/ );
			if ( -d $dir && -f "/etc/mtree/$file" ) {
				print ( STDERR "\tRunning cmd $mtree -f /etc/mtree/$file -x -p $dir -e\n") if ( $opt_d );
				open (fp, "$mtree -f /etc/mtree/$file -x -p $dir -e|") || die "Could not run $mtree";
				while (<fp>) {
					chop;
					if ( length $_ > 1 ) {
						if ( $_ =~ /missing:/o ) {
							$dir2 = $_;
							$dir2 =~ s,missing: ./,,o;
							$dir2 = "$dir/$dir2" if ( $dir ne '/');
							$dir2 = "/$dir2" if ( $dir eq '/');
							$perms{$dir2} = "file not found";
						} elsif ( $_ =~ /.*:/o ) {
							$dir2 = $_;
							$dir2 =~ s/:.*$//go;
							$dir2 = "$dir/$dir2" if ( $dir ne '/');
							$dir2 = "/$dir2" if ( $dir eq '/');
							$perms{$dir2} = 1;
						} else {
							$perms{$dir2}=($perms{$dir2}==1)? "$_":
								$perms{$dir2} . $_;
							$perms{$dir2} =~ s/  / /go;
						}
					}
				}
				close fp;
			}
		}
		closedir DIR;
	}
}

sub run_pkgchk {
	my ($dir);
	print ( STDERR "In run_pkgchk\n") if ( $opt_d );

	print ( STDERR "\tRunning cmd $pkgchk -n -q\n") if ( $opt_d );
	if (open (fp, "$pkgchk -n -q 2>&1 |")) {
		while (<fp>) {
			chop;
			if ( $_ =~ /ERROR: /o ) {
				$dir = $_;
				$dir =~ s/^ERROR: //go;
				$perms{$dir} = 1;
			} elsif ( length ($dir) > 1 ) {
				s/file //go;
				s/permissions /mode /o;
				s/group name /group /o;
				s/\> expected \</:/o;
				s/ actual//o;
				s/\</(/o;
				s/\>/)/o;
				$perms{$dir}=($perms{$dir}==1)? "$_":
					$perms{$dir} . $_;
				$perms{$dir} =~ s/  / /go;
			} else {
				print ( STDERR "\tDo not know what to do with $_\n\tdir = $dir\n") if ( $opt_d );
			}
		}
		close fp;
	} else {
		print ( STDERR "ERROR: Running cmd $pkgchk -n -q\n") if ( $opt_d );
	}
}

sub run_showfiles {
	print ( STDERR "In run_showfiles\n") if ( $opt_d );

	my ($type, $cursum, $sum, $curblock, $bytes, $package, $dummy, $name);
     my ($dev,$ino,$mode,$nlink,$uid,$gid,%checked,$rdev,$curbytes);
	my (%missingfile, %badbytes, %badsum, %badtype);
     open (fp, "showfiles -A -M |") || die "Could not run showfiles";
     while (<fp>) {
          ($type, $sum, $bytes, $package, $dummy, $name) =
               split (/[\s\t]+/o, $_);

          $name = $dummy if ( !length ($name));

          # skip the configuration files, they should have changed
          next if ( $dummy eq 'c' );

          if ( ! -e $name ) {
               $missingfile{$name} = 1;
               next;
          }

          ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$curbytes) = lstat($name);

		# need to use /bin/sum -r because that is what showfiles uses.
          if ( $type eq 'f' && $sum )  {
               ($cursum, $curblock, $dummy) =
                    split (/[\s\t]+/o, `/bin/sum -r $name`);

               # Any file definitions from patch packages rule.
               next if ( defined $checked{$name} && $package !~ /patch/o );

               if ( $sum != $cursum ) {
                    $badsum{$name} = $sum;
                    $badcursum{$name} = $cursum;
               } elsif ( defined $badsum{$name} ) {
                    delete $badsum{$name};
                    delete $badsum{$name};
               }

               if ( $bytes != $curbytes ) {
                    $badbytes{$name} = $bytes;
                    $badcurbytes{$name} = $curbytes;
               } elsif ( defined $badbytes{$name} ) {
                    delete $badbytes{$name};
               }
          }

          $badtype{$name} = $mode if ($type eq 'F' && !($mode & 0x1000));
          $badtype{$name} = $mode if ($type eq 'c' && !($mode & 0x2000));
          $badtype{$name} = $mode if ($type eq 'd' && !($mode & 0x4000));
          $badtype{$name} = $mode if ($type eq 'b' && !($mode & 0x6000));
          $badtype{$name} = $mode if ($type eq 'f' && !($mode & 0x8000));
          $badtype{$name} = $mode if ($type eq 'l' && !($mode & 0xA000));
          $badtype{$name} = $mode if ($type eq 's' && !($mode & 0xC000));

          # Now add it to the list of files checked
          $checked{$name} = 1;
     }
     close fp;

	# Now process the info so we can display it correctly
     my ($dir);
     my (@f_keys) = keys %missingfile;
     my (@s_keys) = keys %badsum;
     my (@t_keys) = keys %badtype;
     my (@keys) = (@f_keys, @s_keys, @t_keys);
     foreach $dir (sort @keys) {
		$perms{$dir} = "$perms{$dir} pathname does not exist" 
			if (defined $missingfile{$dir});

          $perms{$dir} = "$perms{$dir} size ($badbytes{$dir}:$badcurbytes{$dir})" 
			if (defined $badbytes{$dir});

          $perms{$dir} = "$perms{$dir} cksum ($badsum{$dir}:$badcursum{$dir})" 
			if (defined $badsum{$dir});

          $perms{$dir} = "$perms{$dir} pathname not correct type" 
			if (defined $badtype{$dir});

		$perms{$dir} =~ s/  / /go;
     }
}

# Check to see what system files are group or world writable
# Uses the system defined checking commands like mtree, pkgchk & showfiles.
sub file_permissions {
	print ( STDERR "In file_permissions\n") if ( $opt_d );

	run_mtree if ( defined $mtree && -d '/etc/mtree' );
	run_pkgchk if ( defined $pkgchk );
	run_showfiles if ( defined $showfiles );

	# First rename yesterday's file if needed.
	if ( -f "$CONFDIR/fileperms.today" ) {
		rename ("$CONFDIR/fileperms.today", "$CONFDIR/fileperms.yesterday");
	} else {
		&create_secure_file ("$CONFDIR/fileperms.yesterday", 0, 0, 0600);
	}
	&create_secure_file ("$CONFDIR/fileperms.today", 0, 0, 0600);

	# Now write out today's info for tomorrow
	open (FP, "> $CONFDIR/fileperms.today") || die "Could not open fileperms.today";
	foreach $i (sort keys %perms) {
		print FP "$i $perms{$i}\n";
	}
	close FP;
}

sub check_conf_file {
	print ( STDERR "In check_conf_file\n") if ( $opt_d );
	my ($file, $wanted_mode, $wanted_uid, $msg) = @_;
	my ($cursum);

	if ( ! -e $file ) {
		die "$msg does not exist" if (! $opt_c );
		&create_secure_file ($file, $wanted_uid, 0, $wanted_mode) 
			if ( $file ne $CONFDIR );
	}

    	my($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($file);

	printf ( STDERR "\t%s mode = %o, uid = %d, gid = %d\n", 
		$file, ($mode & 00777), $uid, $gid) if ( $opt_d );

	die "Permissions, owner or group not correct on $msg" 
		if (($mode & 00777) != $wanted_mode || 
			$uid != $wanted_uid || 
			$gid != 0 );

	if ( $file ne "$CONFDIR/.digest" && 
		$file ne $CONFDIR && 
		length ($chksum{$file}) > 1 ) {
		if ( $digestfunc eq 'sum' ) {
			grep (/^$format$/ && ($cursum=$1), &{$digestfunc}($file));
		} else {
			grep (/^$format$/ && ($cursum=$2), &{$digestfunc}($file));
		}

		printf ( STDERR "\t$file checksum <%s> actual <%s> expected\n", 
			$cursum, $chksum{$file}) if ( $opt_d );

		die "$msg has bad checksum" if ( $chksum{$file} != $cursum );
	}
}

sub configure {
	my ($version);
	$digestfunc = 'SHA';
	$sha_ver = 'SHA1';
	$format = "$sha_ver\\((.*)\\) = (.*)";

	print ( STDERR "In configure\n") if ( $opt_d );

	# Define the configuration directory and check to make sure nothing
	# has been mucked around with it.
	$CONFDIR="/.security";
	&check_conf_file ($CONFDIR, 00700, 0, "config dir");

	# Now check the digest file to make sure all the other files are ok.
	# This idea came from the FreeBSD ports collection
	&check_conf_file ("$CONFDIR/.digest", 00600, 0, "digest file");

	# Now read in the digest data
	open (fp, "$CONFDIR/.digest") || die "Could not open digest file";
	while (<fp>) {
		grep (/^$format$/ && ($chksum{$1}=$2), $_);
	}
	close fp;
	unlink ( "$CONFDIR/.digest" );

	# Now check if the configfile is ok and then read in.
	&check_conf_file ("$CONFDIR/security.cf", 00400, 0, "config file");

	open (fp, "$CONFDIR/security.cf") || die "Could not read config file";
	while (<fp>) {
		chop;
		$ignoredir = substr ($_, index ($_, ' ')+1) 
			if (/^ignoredir /o);
		@ignoredevdir = split (/[ \t]+/o, substr ($_, index ($_, ' ')+1)) 
			if (/^ignoredevdir /o);
		@ignorerootports = split (/[ \t]+/o, substr ($_, index ($_, ' ')+1)) 
			if (/^ignorerootports /o);
		$ignoreuid = substr ($_, index ($_, ' ')+1) 
			if (/^ignoreuid /o);
		$ignorepasswd = substr ($_, index ($_, ' ')+1)
			if (/^ignorepasswd /o);
		$intruderfiles = substr ($_, index ($_, ' ')+1)
			if (/^intruderfiles /o);
	}

	printf STDERR "\tignoredir = $ignoredir\n" if ( $opt_d );
	for (my ($i) = 0; $i <= $#ignoredevdir; $i++) {
		printf STDERR "\tignoredevdir = $ignoredevdir[$i]\n" if ( $opt_d );
	}
	for (my ($i) = 0; $i <= $#ignorerootports; $i++) {
		printf STDERR "\tignorerootports = $ignorerootports[$i]\n" if ( $opt_d );
	}
	printf STDERR "\tignoreuid = $ignoreuid\n" if ( $opt_d );
	printf STDERR "\tignorepasswd = $ignorepasswd\n" if ( $opt_d );
	printf STDERR "\tintruderfiles = $intruderfiles\n" if ( $opt_d );
	close fp;

	# Check if this script itself is ok
	&check_conf_file ($0, 00700, 0, "security script");

	# Now check the setuid.* files are ok
	&check_conf_file ("$CONFDIR/setuid.today", 00600, 0, "setuid config file");
	&check_conf_file ("$CONFDIR/setuid.yesterday", 00600, 0, "old setuid config file");

	# Now check the fileperms.* files are ok
	&check_conf_file ("$CONFDIR/fileperms.today", 00600, 0, "file permsission config file");
	&check_conf_file ("$CONFDIR/fileperms.yesterday", 00600, 0, "old file permsission config file");

	# Mail report information
	$date = strftime ( "%m/%d/%Y", localtime(time()));
	($host) = split('\.', hostname(), 9999);
	open (fp, $0) || die "Could not open myself $0";
	grep (/@\(#\) .[^ ]* ([^ ]*)/ && ($version=$1), <fp>);
	close fp;

	# Find the Mail command to send out report
	$mailcmd = '/usr/bin/Mail';
	$mailcmd = '/usr/sbin/Mail' if ( -f '/usr/sbin/Mail' );;
	$mailcmd = '/usr/bin/mailx' if ( -f '/usr/bin/mailx' );
	$mailcmd = '/usr/ucb/Mail' if ( -f '/usr/ucb/Mail' );
	$mailargs = "-s \"$host security ($version) report for $date\"";
	$mailargs = "-v $mailargs" if ( $opt_d );
	$address = 'sysnotify' if ( ! defined( $address ));

	# Find the files we need to run
	$passwd = '/etc/passwd';
	$passwd = '/etc/master.passwd' if ( -f '/etc/master.passwd' && $^O != 'darwin');
	printf STDERR "\tpasswd = $passwd\n" if ( $opt_d );
	$shadow = '/etc/shadow' if ( -f '/etc/shadow');
	$showmount = '/usr/etc/showmount';
	$showmount = '/usr/sbin/showmount' if ( -f '/usr/sbin/showmount' );
	$mtree = '/usr/sbin/mtree' if ( -f '/usr/sbin/mtree' );
	$pkgchk = '/usr/sbin/pkgchk' if ( -f '/usr/sbin/pkgchk' );
	$showfiles = '/usr/sbin/showfiles' if ( -f '/usr/sbin/showfiles' );
	$diff = '/usr/bin/diff';

	# Find the sum command
	$sum = '/bin/sum -r' if ( -f '/bin/sum' );
	$sum = '/usr/bin/sum -r' if ( -f '/usr/bin/sum' );

	# Find the directory where the loadable kernel modules are
	$sysctl = '/usr/sbin/sysctl' if ( -f '/usr/sbin/sysctl' );
	$sysctl = '/sbin/sysctl' if ( -f '/sbin/sysctl' );
	if ( -f $sysctl ) {
		chop ($kerneldir=`$sysctl -n kern.module_path 2> /dev/null`);
		$kerneldir =~ s,^/;,^,o;
		$kerneldir =~ s/;/|^/go;
	} else {
		foreach $dir ('/lkm', '/kernel/drv', '/System/Library/Extensions') {
			$kerneldir .= '|' if ( defined $kerneldir );
			$kerneldir .= '^' . $dir if ( -d $dir );
		}
	}
	printf STDERR "\tKerneldir = $kerneldir\n" if ( $opt_d );

	$loginfile = '/etc/default/login' if ( -f '/etc/default/login' );
	$accessfile = '/etc/login.access' if ( -f '/etc/login.access' );
	$securefile = '/etc/ttytab' if ( -f '/etc/ttytab' );
	printf STDERR "\tloginfile = $loginfile\n" if ( $opt_d );
	printf STDERR "\taccessfile = $accessfile\n" if ( $opt_d );
	printf STDERR "\tsecurefile = $securefile\n" if ( $opt_d );
}

# Check for files that have no defined owner or group.
# Check to see any character or block special devices not in /dev.  Again 
# from HP-UX System Security manual
#
# Added lib*.so* checks and $kerneldir based on the SANS98 Conference
sub wanted {
	if (index (" $ignoredir ", " $File::Find::name ") >= 0) {
		$File::Find::prune = 1;
		printf STDERR "\tPruning Tree at $File::Find::name\n" if ( $opt_d );
		return;
	} 

	my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$curbytes) = lstat($_);

	if ( -d _ ) {

		my($bsize, $frsize, $blocks, $bfree, $bavail,
			$files, $ffree, $favail, $fsid, $basetype, $flag,
			$namemax, $fstr)=statvfs($_);

		if ( $basetype =~ /(nfs)|(autofs)/o ) {
			$File::Find::prune = 1;
			printf STDERR "\tPruning Tree at $File::Find::name ($basetype)\n" if ( $opt_d );
			return;
		}

	} elsif ( -f _ ) {

		# If the file is set uid, set gid, a shared library or a kernel
		# module, we want to track it's checksum.
		if (($mode & &S_ISUID()) || ($mode & &S_ISGID()) || 
			/^lib.*\.so.*$/o || 
			( defined $kerneldir && $File::Find::name =~ /$kerneldir/ )) {

			$setuid{$File::Find::name} = 
				&{$digestfunc}($File::Find::name);

			# If the file is world writeable, then flag this case too
			$setuidperms{$File::Find::name} = 1
				if ( $mode & &S_IWOTH() );
		}

	# If the file is a block or character special device that is not in
	# /dev, then flag this case.
	} elsif ( -b _ || -c _ ) {
		my ($res) = 1;
		for (my ($i)= 0; $i <= $#ignoredevdir; $i++) {
			$res = 0 if ( $File::Find::name =~ /^$ignoredevdir[$i]/ );
		}
		$devs{$File::Find::name} = 1 if ($res);
	}

	# If the file has no known owner or group, add it to that list
     $nouids{$File::Find::name} = 1
		if ( !defined $gid{$gid} || !defined $uid{$uid} );
}

sub mail_results {
	$file_open=0;
	print ( STDERR "In mail_results\n") if ( $opt_d );

	if ( ! $opt_f ) {
		open (DIFF, "$diff -b $CONFDIR/setuid.yesterday $CONFDIR/setuid.today|");
		my ($cnt) = 0;
		while (<DIFF>) {
			&check_fp();
			if (! $cnt ) {
				print fp "checking setuid files and devices:\n$host setuid/device diffs:\n";
				$cnt=1;
			}
			print fp $_;
		}
		close DIFF;
	}

	&print_data ("\n\nchecking for world writable setuid/setgid programs:\n", %setuidperms);
	&print_data ("\n\nchecking for uids of 0:\n", %uid0);
	&print_data ("\n\nchecking for accounts with no passwords:\n", %nopass);
	&print_data ("\n\nchecking for duplicate uids:\n", %dupuid);
	&print_data ("\n\nchecking for world exported filesystems:\n", %exports);
	&print_data ("\n\nchecking for block & character special devices not in /dev:\n", %devs);
	&print_data ("\n\nchecking for files with no owner or group:\n", %nouids);

	if ( ! $opt_m ) {
		open (DIFF, "$diff -b $CONFDIR/fileperms.yesterday $CONFDIR/fileperms.today|");
		my ($cnt) = 0;
		while (<DIFF>) {
			&check_fp();
			if (! $cnt ) {
				print fp "\n\nchecking file permission problems:\n$host file permission diffs:\n";
				$cnt=1;
			}
			print fp $_;
		}
		close DIFF;
	}

	&print_data ("\n\nchecking for known intruder files:\n", %badfile);

	if ( ! $rootlogin ) {
		&check_fp();
		print fp "\n\nRoot login allowed on all ports\n";
	}

	&print_data ("\n\nchecking for ports root is allowed to login to:\n", %ttys);

	close fp;
}

sub check_fp {
	if ( !$file_open ) {
		if ( $address =~ /^\//o ) {
			open (fp, ">$address" ) 
				|| die "Could not open $address";
		} else { 
			open (fp, "|$mailcmd $mailargs $address" ) 
				|| die "Could not execute $mailcmd";
		}
		$file_open = 1;
	}
}

sub print_data {
	print ( STDERR "In print_data\n") if ( $opt_d );
	my ($header,%ary) = @_;
	my ($cnt) = 0;
	foreach $i (sort keys %ary) {
		&check_fp();
		if (! $cnt ) {
			print fp $header;
			$cnt=1;
		}
		print (fp $i);
		print (fp " $ary{$i}") if $ary{$i} != 1;
		print (fp "\n");
	}
}

sub load_YPpasswd {
	my ($res, $nm, $val, $pw, $uid, $dummy);

#	for (($res, $nm, $val) = Net::NIS::yp_first($YP, 'passwd.byname'); 
#          $res == 0;
#          ($res, $nm, $val) = Net::NIS::yp_next($YP, 'passwd.byname', $nm)) {
#		($nm, $pw, $uid, $dummy) = split (/:/, $val, 4);
#		$uid{$nm} = $uid{$uid} = $uid;
#	}
}
sub load_YPgroup {
	my ($res, $nm, $val, $pw, $gid, $dummy);

#	for (($res, $nm, $val) = Net::NIS::yp_first($YP, 'group.byname'); 
#          $res == 0;
#		($res,$nm,$val) = Net::NIS::yp_next($YP,'group.byname',$nm)) {
#          ($nm, $pw, $gid, $dummy) = split (/:/, $val, 4);
#          $gid{$nm} = $gid{$gid} = $gid;
#	}
}

sub proccess_options {
	my ($opts) = 'cdefhiM:mpry';
	my ($loadedyp) = 0;
	do getopts ($opts);

	if ( $opt_h ) {
		# shorten the program name
		$0 = substr ($0, rindex ($0, "/") + 1);

		print "Usage: $0 -[$opts]\n";
		print "\t-c        Create dummy config files\n";
		print "\t-d        Debug mode\n";
		print "\t-e        Skip the exported file system check\n";
		print "\t-f        Skip the find and all its check\n";
		print "\t-h        This help screen\n";
		print "\t-i        Skip the intruder file check\n";
		print "\t-r        Skip root login check\n";
		print "\t-M <addr> Mail report to <addr>, instead of \"sysnotify\"\n";
		print "\t-m        Skip the file system sanity checks (mtree, pkgchk, showfiles)\n";
		print "\t-p        Skip the passwd check\n";
		print "\t-y        Skip loading the YP maps for passwd/group\n";
		exit 0;
	}
	$address = $opt_M if ( $opt_M );

	# Init the YP stuff if needed
#	if ( ! $opt_y ) {
#		use Net::NIS;
#		$YP = Net::NIS::yp_get_default_domain();
#	}

	# Load up some tables for the find to run
	while (($nm, $pw, $uid) = getpwent) {
		if ( $nm eq '+' && ! $opt_y ) {
			$loadedyp=1;
			&load_YPpasswd;
		} else {
			$uid{$nm} = $uid{$uid} = $uid;
		}
	}
	&load_YPpasswd if ( $loadedyp != 1  && -f '/etc/defaultdomain' );

	while (($nm, $pw, $gid) = getgrent) {
		if ( $nm eq '+' && ! $opt_y ) {
			$loadedyp=2;
			&load_YPgroup;
		} else {
			$gid{$nm} = $gid{$gid} = $gid;
		}
	}
	&load_YPgroup if ( $loadedyp != 2  && -f '/etc/defaultdomain' );

	$| = 1 if ( $opt_d );
}

sub update_sum {
	my ($file, $sum);
	my (@files) = ( $0, 
			"$CONFDIR/security.cf",
			"$CONFDIR/setuid.today",
			"$CONFDIR/setuid.yesterday",
			"$CONFDIR/fileperms.today",
			"$CONFDIR/fileperms.yesterday" );
	print ( STDERR "In update_sum\n") if ( $opt_d );

	&create_secure_file ("$CONFDIR/.digest", 0, 0, 0600);
	open (FP, "> $CONFDIR/.digest") || die "Could not open digest file";

	for $file (@files) {
		$sum = &{$digestfunc}($file);
		print ( STDERR "\tSum = $sum" ) if ( $opt_d );
		print FP $sum || die "Error saving digest digest";
	}
	close FP;
}

sub check_root_login {
	local *FP;
	$rootlogin = -1;

	print ( STDERR "In check_root_login\n") if ( $opt_d );
	if ( $loginfile ) {
		open (FP, $loginfile) || die "Could not open login file $loginfile";
		$rootlogin=grep (/^CONSOLE/o, <FP>);
		close FP;
	} elsif ( $securefile ) {
		open (FP, $securefile) || die "Could not open login file $securefile";
		grep (/^([^ \t#]*).*on[ \t]*secure$/o && ($ttys{$1}=1), <FP>);
		close FP;
	} elsif ( $accessfile ) {
		open (FP, $accessfile) || die "Could not open login file $accessfile";
		while (<FP>) {
			my (@a) = split (/:/o, $_) if ( $_ !~ /^#/o );
			if ( $a[1] eq "root" ) {
				if ( $a[0] eq "+" ) {
					if ( $a[2] =~ /ALL$/o ) {
						$rootlogin=0;
					} else {
						my (@b) = split (/[ \t]+/o, $a[2]);
						for (my ($i) = 0; $i < $#b; $i++) {
							$ttys{$b[$i]} = 1 if ( $b[$i] ne 'ALL' );
						}
					}
				} else {
					if ( $a[2] =~ /ALL$/o ) {
						$rootlogin=1;
					} else {
						my (@b) = split (/[ \t]+/o, $a[2]);
						for (my ($i) = 0; $i < $#b; $i++) {
							$ttys{$b[$i]} = 1 
								if ( $b[$i] ne 'ALL' && 
									$b[$i] ne 'EXCEPT' );
						}
					}
				}
			}
		}
		close FP;
	}

	# Now remove the preconfigured root ports
	for (my ($i) = 0; $i <= $#ignorerootports; $i++) {
		if ( defined $ttys{$ignorerootports[$i]} ) {
			printf ( STDERR "unsetting $ignorerootports[$i] in ttys array\n") 
				if ( $opt_d );
			delete $ttys{$ignorerootports[$i]};
		}
	}
	
}

&proccess_options;
&configure;
if ( ! $opt_f ) {
	find(\&wanted, '/');
	&check_setuid;
}
&file_permissions if ( ! $opt_m );
&check_intruder_files if ( ! $opt_i ); 
&check_root_login if ( ! $opt_r ); 
&check_exports if ( ! $opt_e ); 
&check_passwd if ( ! $opt_p );
&update_sum;
&mail_results;

exit 0
